"""This module contains code for running a specific test set"""
# pylint: disable=too-few-public-methods
from typing import List, Set

from _pytest.main import Session
from _pytest.nodes import Item

from pytest_rts.pytest.fake_item import FakeItem
from pytest_rts.utils.common import filter_pytest_items


class RunnerPlugin:
    """Plugin class for pytest"""

    def __init__(self, existing_tests: Set[str]) -> None:
        """Set existing tests"""
        self.existing_tests = existing_tests

    def pytest_collection_modifyitems(
        self,
        session: Session,
        config,  # pylint: disable=unused-argument
        items: List[Item],
    ) -> None:
        """Select only specific tests for running"""
        original_length = len(items)
        items[:] = filter_pytest_items(items, self.existing_tests)
        session.config.hook.pytest_deselected(
            items=([FakeItem(session.config)] * (original_length - len(items)))
        )



"""Code for pytest-rts plugin logic"""
import os

import pytest
from _pytest.config import Config
from _pytest.config.argparsing import Parser

from pytest_rts.pytest.runner_plugin import RunnerPlugin
from pytest_rts.utils.common import get_existing_tests


def pytest_addoption(parser: Parser) -> None:
    """Register pytest flags"""
    group = parser.getgroup("pytest-rts")
    group.addoption("--rts", action="store_true", default=False, help="Run pytest-rts")
    group.addoption(
        "--rts-coverage-db",
        action="store",
        default="",
        help="Coverage file pytest-rts",
    )


def pytest_configure(config: Config) -> None:
    """Register RTS plugins based on state"""
    if not config.option.rts:
        return

    if not config.option.rts_coverage_db:
        pytest.exit("No coverage file provided", 2)

    if not os.path.exists(config.option.rts_coverage_db):
        pytest.exit("Provided coverage file does not exist", 2)

    existing_tests = get_existing_tests(config.option.rts_coverage_db)
    config.pluginmanager.register(RunnerPlugin(existing_tests))



"""Tests for common utility functions"""
from typing import cast

import pytest
from _pytest.nodes import Item

from pytest_rts.utils.common import filter_pytest_items, strip_pytest_cov_testname


@pytest.mark.parametrize(
    "testname, expected",
    [
        (
            "tests/test_example.py::test_methods_var_inheritance|setup",
            "tests/test_example.py::test_methods_var_inheritance",
        ),
        (
            "tests/test_example2.py::TestJSON::test_jsonify_basic_types[0]|teardown",
            "tests/test_example2.py::TestJSON::test_jsonify_basic_types[0]",
        ),
        (
            "tests/test_example3.py::test_session_ip_warning|run",
            "tests/test_example3.py::test_session_ip_warning",
        ),
    ],
)
def test_strip_pytest_cov_testname(testname: str, expected: str) -> None:
    """Test pytest-cov testname stripping to actual pytest item.nodeid strings"""
    assert strip_pytest_cov_testname(testname) == expected


def test_filter_pytest_items() -> None:
    """Test for filtering the test set
    based on given existing tests
    """

    class FakePytestItem:  # pylint: disable=too-few-public-methods
        """Class that has nodeid field like a pytest item
        and a method get_closest_marker which returns something
        """

        def __init__(self, name: str, nodeid: str, markername: str) -> None:
            """Set test name, identifier nodeid and marker name"""
            self.name = name
            self.nodeid = nodeid
            self.markername = markername

        def get_closest_marker(self, markername: str) -> bool:
            """Return if item is marked as skipped"""
            if self.markername == markername:
                return True
            return False

    collected_items = [
        cast(Item, FakePytestItem("test1", "test_func_1", "")),
        cast(Item, FakePytestItem("test2", "test_func_2", "skip")),
        cast(Item, FakePytestItem("test3", "test_func_3", "")),
        cast(Item, FakePytestItem("test4", "test_func_4", "skipif")),
        cast(Item, FakePytestItem("test5", "test_func_5", "")),
    ]
    existing_tests = {
        "test_func_1",
        "test_func_3",
    }

    filtered_items = filter_pytest_items(collected_items, existing_tests)

    assert len(filtered_items) == 1
    assert filtered_items[0].nodeid == "test_func_5"



class Car:

    def __init__(self,seats,speed,passengers):
        self.seats = seats
        self.speed = speed
        self.passengers = passengers

    def get_speed(self):
        return self.speed

    def accelerate(self, amount):
        self.speed = self.speed + amount

    def get_passengers(self):
        return self.passengers

    def add_passenger(self):
        if self.passengers + 1 <= self.seats:
            self.passengers = self.passengers + 1
        
    def remove_passenger(self):
        if self.passengers > 0:
            self.passengers = self.passengers - 1


import pytest
import sys
sys.path.append(".")
from src.shop import Shop

def test_normal_shop_purchase():
    shop = Shop(5,10,0)
    shop.buy_item()
    shop.buy_item()

    item_price = shop.get_item_price()
    assert shop.get_items() == 8
    assert shop.get_money() == item_price * 2

def test_normal_shop_purchase2():
    shop = Shop(5,10,0)
    shop.buy_item()
    shop.buy_item()
    shop.buy_item()
    shop.buy_item()

    item_price = shop.get_item_price()

    assert shop.get_items() == 6
    assert shop.get_money() == item_price * 4

def test_empty_shop_purchase():
    shop = Shop(5,0,0)
    shop.buy_item()
    assert shop.get_money() == 0

	





"""Integration tests for pytest-rts"""
import os

from _pytest.pytester import Testdir


def test_only_new_functions_are_ran(testdir: Testdir) -> None:
    """Test case for running pytest-rts when new tests are added
    and changes are not committed.
    """
    # copy the helper project
    testdir.copy_example(".")

    # run pytest pytest-cov to produce mapping
    # subprocess used to not mess up with coverage for this test suite
    testdir.runpytest_subprocess("--cov=.", "--cov-context=test")

    # rename .coverage file produced by pytest-cov
    os.rename(".coverage", "rts-coverage.1234")

    # add a new test method to an existing file
    os.rename("changes/test_car/add_test_passengers.txt", "tests/test_car.py")

    # add a new test file with a test in it
    testdir.makepyfile(
        """
        def test_empty():
            assert 1 == 1
    """
    )

    # run pytest-rts to only run new tests
    result = testdir.runpytest("--rts", "--rts-coverage-db=rts-coverage.1234")
    result.assert_outcomes(passed=2, failed=0, errors=0)



"""This module contains fuctions for test selecting operations"""
from typing import List, Set

from coverage import CoverageData
from _pytest.nodes import Item


def filter_pytest_items(
    pytest_items: List[Item], existing_tests: Set[str]
) -> List[Item]:
    """Select pytest items if they are new and not marked as skipped"""
    return list(
        filter(
            lambda item: item.nodeid not in existing_tests
            and not item.get_closest_marker("skipif")
            and not item.get_closest_marker("skip"),
            pytest_items,
        )
    )


def get_existing_tests(coverage_file_path: str) -> Set[str]:
    """Read all the test function names from the coverage file.
    pytest-cov creates the coverage file and adds a section at the
    end of each testname which need to be stripped.
    """
    coverage_data = CoverageData(coverage_file_path)
    coverage_data.read()
    return {
        strip_pytest_cov_testname(testname)
        for testname in coverage_data.measured_contexts()
    }


def strip_pytest_cov_testname(testname: str) -> str:
    """Strip ends of pytest-cov produced testnames"""
    if testname.endswith("|teardown"):
        return testname[:-9]
    if testname.endswith("|setup"):
        return testname[:-6]
    if testname.endswith("|run"):
        return testname[:-4]
    return testname
